<footer class="footer">
    <div class="container-fluid">
        <nav class="pull-left">
          <ul>
            <li><a href="http://www.servisys.com">Servisys inc.</a></li>
            <li><a href="https://pythoninthebuilding.wordpress.com">Blog</a></li>
            <li><a href="https://github.com/JoelBender/bacpypes">BACpypes</a></li>
          </ul>
        </nav>
        <div class="copyright pull-right">
        &copy; <script>document.write(new Date().getFullYear())</script>, Web design made with <i class="fa fa-heart heart"></i> by <a href="http://www.creative-tim.com">Creative Tim</a>
        </div>
    </div>
</footer>
