{% extends "bootstrap/base.html" %}

{% block head %}
{{ super() }}
<link rel="apple-touch-icon" sizes="76x76" href="static/assets/img/apple-icon.png">
<link rel="icon" type="image/png" sizes="96x96" href="static/assets/img/favicon.png">
{% endblock %}

{% block metas %}
{{ super() }}
<meta charset="utf-8" />
<meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1" />
<meta content='width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0' name='viewport' />
<meta name="viewport" content="width=device-width" />
{% endblock %}

{% block title %}
BAC0 - A Python App for BACnet
{% endblock %}
	
{% block styles %}
{{ super() }}
<!-- Bootstrap core CSS     -->
<link href="static/assets/css/bootstrap.min.css" rel="stylesheet" />
<!-- Animation library for notifications   -->
<link href="static/assets/css/animate.min.css" rel="stylesheet"/>
<!--  Paper Dashboard core CSS    -->
<link href="static/assets/css/paper-dashboard.css" rel="stylesheet"/>
<!--  CSS for Demo Purpose, don't include it in your project     -->
<link href="static/assets/css/demo.css" rel="stylesheet" />
<!--  Fonts and icons     -->
<link href="http://maxcdn.bootstrapcdn.com/font-awesome/latest/css/font-awesome.min.css" rel="stylesheet">
<link href='https://fonts.googleapis.com/css?family=Muli:400,300' rel='stylesheet' type='text/css'>
<link href="static/assets/css/themify-icons.css" rel="stylesheet">
{% endblock %}

{% block scripts %}
{{ super() }}
<!--   Core JS Files   -->
<script src="static/assets/js/jquery-1.10.2.js" type="text/javascript">
<script src="static/assets/js/bootstrap.min.js" type="text/javascript">
<!--  Checkbox, Radio & Switch Plugins -->
<script src="static/assets/js/bootstrap-checkbox-radio.js">
<!--  Charts Plugin -->
<script src="static/assets/js/chartist.min.js">
<!--  Notifications Plugin    -->
<script src="static/assets/js/bootstrap-notify.js">
<!--  Google Maps Plugin    -->
<script type="text/javascript" src="https://maps.googleapis.com/maps/api/js">
<!-- Paper Dashboard Core javascript and methods for Demo purpose -->
<script src="static/assets/js/paper-dashboard.js">
<!-- Paper Dashboard DEMO methods, don't include it in your project! -->
<script src="static/assets/js/demo.js">
<script src="static/assets/js/BAC0_script.js">
{% endblock %}

