{% extends "navbar.html" %}
{% block content %}
<div class="wrapper">
{{sidebar|safe}}

  <div class="main-panel">

  <div class="content">
    <div class="container-fluid">
      <div class="row">
        {{bokeh_script|safe}}
      </div>
    </div>
  </div>

  {% include "footer.html" %}
  </div>	
</div>
{% endblock %}
