{% extends "base.html" %}
{% block navbar %}
<nav class="navbar navbar-default">
  <div class="container-fluid">
    <div class="navbar-header">
    <button type="button" class="navbar-toggle">
      <span class="sr-only">Toggle navigation</span>
      <span class="icon-bar bar1"></span>
      <span class="icon-bar bar2"></span>
      <span class="icon-bar bar3"></span>
    </button>
    <a class="navbar-brand" href="#">Dashboard</a>
    </div>

    <div class="collapse navbar-collapse">
    <ul class="nav navbar-nav navbar-right">
      <li><a href="" class="dropdown-toggle" data-toggle="dropdown">							           <i class="ti-server"></i>
      <p id="discover">Discover</p>
      </a>
      </li>

      <li class="dropdown"><a href="#" class="dropdown-toggle" data-toggle="dropdown">
      <i class="ti-bell"></i>
      <p class="notification"></p>
      <p>Notifications</p>
      <b class="caret"></b>
      </a>
      {{notifications|safe}}
      </li>
    </ul>
    </div>

  </div>
</nav>
{% endblock %}
