{% extends "navbar.html" %}
{% block content %}
<div class="wrapper">

{{sidebar|safe}}

    <div class="main-panel">
        <div class="content">
            <div class="container-fluid">
                <div class="row">
                    {{card_number_of_devices|safe}}
                    
                    {{card_number_of_mstp_networks|safe}}

                    {{card_number_of_trends|safe}}
                    
                </div>

                <div class="row">
                    <div class="col-md-6">
                        <div class="card">
                            <div class="header">
                                <h4 class="title">Networks and devices Statistics</h4>
                                <p class="category">Distribution over networks</p>
                            </div>
                            <div class="content">
                                {{bokeh_pie|safe}}
                            </div>
                        </div>
                    </div>
                    
                </div>
            </div>
        </div>

        {% include "footer.html" %}


    </div>
</div>
{% endblock %}


